/* $Id: mesh.c,v 1.25 1998/03/16 05:45:05 svmcguir Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "path.h"
#include "option.h"
#include "mesh.h"
#include "util.h"

void meshDataAllocate(meshData *md){
  /* Allocate memory for md.valueIfHeld and md.value
   */

  allocate2dDouble(&(md->valueIfHeld), md->option.pData.times, 
                   md->option.pData.meshPaths);
  allocate2dDouble(&(md->newValueIfHeld),  md->option.pData.times,
                   md->option.pData.meshPaths);
  allocate2dDouble(&(md->value), md->option.pData.times,
                   md->option.pData.meshPaths);
  allocate2dDouble(&(md->newValue), md->option.pData.times,
                   md->option.pData.meshPaths);
  allocate2dDouble(&(md->tp), md->option.pData.meshPaths, md->meshPaths);
  allocate2dDouble(&(md->distribution), md->option.pData.times, md->meshPaths);
  allocate1dDouble(&(md->sum), md->meshPaths);
  allocate1dDouble(&(md->weightMeasure1), md->option.pData.times);
  allocate1dDouble(&(md->weightMeasure2), md->option.pData.times);
  allocate1dDouble(&(md->weightMeasure3), md->option.pData.times);
}

void meshSetFinalValue(meshData *md){
  /* Set the values on the final timeslice to the payoffs or 0, whichever
   * is greater.
   */

  int path,time;

  time = md->option.pData.times-1;

  for(path=0;path<md->option.pData.meshPaths;path++){
    md->value[time][path] = dMax(md->option.payoff[time][path],0);
    md->valueIfHeld[time][path] = 0;
    md->newValue[time][path] = md->value[time][path];
    md->newValueIfHeld[time][path] = md->valueIfHeld[time][path];
  }
}

void meshDataFill(meshData *md){
  /* Working backwards from the next to last timeslice (last is handled 
   * by meshSetFinalValue) calculate the values of the option if it is held
   * (equation 6).  Then set the value of the option to the greater of the 
   * value if held and the payoff.  The first timeslice is left for 
   * meshMeshEstimate().
   */

  int time, path, fromPath, toPath, basePath,underlier;
  double weightSum, weightSqSum, weight;


  for(time=md->option.pData.times-2; time>=0; time--){
    md->weightMeasure1[time] = 0;
    /* Get all the transition probabilities at once */
    for(fromPath=0; fromPath < md->meshPaths; fromPath++){
      for(toPath=0; toPath < md->meshPaths; toPath++){
        md->tp[fromPath][toPath] =
          pathTransitionProb(&(md->option.pData), time, fromPath, toPath);
      }
    }

    /* The denominators of the weights are sums of tp[fromPath][toPath]
     * over fromPath. 
     */
    for(toPath=0; toPath<md->meshPaths; toPath++){
      md->distribution[time][toPath] = 0;
      for(fromPath=0; fromPath<md->meshPaths; 
          fromPath+=md->option.pData.clusterSize){
        md->distribution[time][toPath] += 
          md->option.pData.clusterSize*md->tp[fromPath][toPath];
      }
    }

    /* Now, weight[fromPath][toPath] = tp[fromPath][toPath]/sum[toPath]*/
    for(fromPath=0; fromPath<md->meshPaths; fromPath++){
      md->valueIfHeld[time][fromPath] = 0;
      weightSum = 0;
      weightSqSum = 0;
      for(toPath=0; toPath<md->meshPaths; toPath++){
        if(md->tp[fromPath][toPath] == 0){
          weight = 0;
        }else{
          weight = md->tp[fromPath][toPath]/md->distribution[time][toPath];
        }
      
        weightSum += weight;
        weightSqSum += weight*weight;
        /* Compute value if held */
        md->valueIfHeld[time][fromPath] += weight*md->value[time+1][toPath];

      }
      if(weightSum != 0){
        md->weightMeasure1[time] += weightSum*weightSum/weightSqSum;
      }
      md->value[time][fromPath] = dMax(md->valueIfHeld[time][fromPath],
                                       md->option.payoff[time][fromPath]);
    }
    md->weightMeasure1[time] /= md->option.pData.meshPaths;
  }
}


double meshPathEstimate(meshData *md){
  int time, underlier, toPath, count;
  double **path, **antiPath, antiGrowth, *payoff, *antiPayoff, *valueIfHeld,
    *antiValueIfHeld, value, antiValue, weight, antiWeight, delta; 

  allocate2dDouble(&path, md->option.pData.times, md->option.pData.underliers);
  allocate2dDouble(&antiPath, md->option.pData.times,
                   md->option.pData.underliers);
  allocate1dDouble(&payoff, md->option.pData.times);
  allocate1dDouble(&antiPayoff, md->option.pData.times);
  allocate1dDouble(&valueIfHeld, md->option.pData.times);
  allocate1dDouble(&antiValueIfHeld, md->option.pData.times);

  value = 0;
  antiValue = 0;

  for(count=1; count<=md->option.pData.paths; count++){

    pathPair(&(md->option.pData),path,antiPath);

    for(time=0; time < md->option.pData.times; time++){
      payoff[time] = exp(-time*md->option.riskFreeRate)*
        optionPayoff(&(md->option), path[time]);
      antiPayoff[time] = exp(-time*md->option.riskFreeRate)*
        optionPayoff(&(md->option), antiPath[time]);
    }
      
    for(time=md->option.pData.times-2; time>=0; time--){
      valueIfHeld[time] = 0;
      antiValueIfHeld[time] = 0;
      for(toPath=0; toPath<md->meshPaths; toPath++){
        weight = transProb(&(md->option.pData),path[time],
                           md->option.pData.price[time+1][toPath])/
          md->distribution[time][toPath];
        antiWeight = transProb(&(md->option.pData),antiPath[time],
                               md->option.pData.price[time+1][toPath])/
          md->distribution[time][toPath];
        valueIfHeld[time]     += weight     * md->value[time+1][toPath];
        antiValueIfHeld[time] += antiWeight * md->value[time+1][toPath];
      }
    }

    for(time=0;time< md->option.pData.times;time++){
      if(payoff[time] > valueIfHeld[time]){
        value += payoff[time];
        break;
      }
    }

    for(time=0;time< md->option.pData.times;time++){
      if(antiPayoff[time] > antiValueIfHeld[time]){
        antiValue += antiPayoff[time];
        break;
      }
    }
  }

  value /= md->option.pData.paths;
  antiValue /= md->option.pData.paths;


  deallocate2dDouble(&path, md->option.pData.times, 
                     md->option.pData.underliers); 
  deallocate2dDouble(&antiPath, md->option.pData.times, 
                     md->option.pData.underliers); 
  deallocate1dDouble(&payoff, md->option.pData.times); 
  deallocate1dDouble(&antiPayoff, md->option.pData.times); 
  deallocate1dDouble(&valueIfHeld, md->option.pData.times); 
  deallocate1dDouble(&antiValueIfHeld, md->option.pData.times); 

  return(0.5*(value+antiValue));
}

      
